Package: prova
Title: Nonparametric Probabilistic-Statistical Variate Analysis with
        Automated Markov-Chain Monte Carlo
Version: 1.0.0
Authors@R: c(
  person("PierGianLuca", "Porta Mana", ,
  "pgl@portamana.org", role = c("aut", "cre", "cph"),
  comment = c(ORCID = "0000-0002-6070-0784")),
  person("Aurora", "Grefsrud", ,
  "agre@hvl.no", role = c("ctb"),
  comment = c(ORCID = "0009-0005-9063-4131")),
  person("Håkon", "Mydland", ,
  "haakon.mydland@gmail.com", role = c("ctb"),
  comment = c(ORCID = "0009-0004-9558-8894")),
  person("Maksim", "Ohvrill", ,
  "maksimohvrill@hotmail.com", role = c("ctb"),
  comment = c(ORCID = "")),
  person("Simen", "Hesthamar Hauge", ,
  "simen@hhnet.no", role = c("ctb"),
  comment = c(ORCID = "0009-0005-6158-2846"))
  )
Description: Calculate posterior joint and conditional probabilities, probability distributions of population frequencies, and information-theoretic measures, by means of Bayesian nonparametric methods. Data imputation is automatic and done in a principled way. Markov-chain Monte Carlo calculations are automatically handled and do not require user supervision. Applications range from statistical estimation and probabilistic hypothesis testing to evidence-based inference and decision making, in a wide range of disciplines from astrophysics to medicine. For more details and examples see for instance Porta Mana et al. (2026) <doi:10.31219/osf.io/8nr56>, Dunson & Bhattacharya (2011) <doi:10.1093/acprof:oso/9780199694587.003.0005>, Lindley & Novick (1981) <doi:10.1214/aos/1176345331>, Bernardo & Smith (2000) <doi:10.1002/9780470316870>, Müller et al. (2015) <doi:10.1007/978-3-319-18968-0>. Requires the packages 'Nimble', 'parallel', 'extraDistr'.
License: AGPL (>= 3)
URL: https://pglpm.github.io/prova/, https://github.com/pglpm/prova/
Encoding: UTF-8
Depends: R (>= 4.5.0)
Imports: extraDistr, parallel
Suggests: nimble (>= 1.4.2), knitr, rmarkdown
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
LazyData: true
NeedsCompilation: no
Packaged: 2026-07-07 12:26:48 UTC; pglpm
Author: PierGianLuca Porta Mana [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-6070-0784>),
  Aurora Grefsrud [ctb] (ORCID: <https://orcid.org/0009-0005-9063-4131>),
  Håkon Mydland [ctb] (ORCID: <https://orcid.org/0009-0004-9558-8894>),
  Maksim Ohvrill [ctb],
  Simen Hesthamar Hauge [ctb] (ORCID:
    <https://orcid.org/0009-0005-6158-2846>)
Maintainer: PierGianLuca Porta Mana <pgl@portamana.org>
Repository: CRAN
Date/Publication: 2026-07-16 12:50:08 UTC
Built: R 4.5.2; ; 2026-07-16 13:01:44 UTC; unix
