# Generated by roxygen2: do not edit by hand

S3method(hist,probability)
S3method(plot,probability)
S3method(print,probability)
export(Pr)
export(flexiplot)
export(learn)
export(metadatatemplate)
export(mutualinfo)
export(plotquantiles)
export(pread.csv)
export(pwrite.csv)
export(qPr)
export(rPr)
export(vrtgrid)
import(grDevices)
import(graphics)
import(parallel)
import(stats)
import(utils)
importFrom(extraDistr,rbern)
importFrom(extraDistr,rcat)
