% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{rowcumsum}
\alias{rowcumsum}
\title{Cumulative sum along first dimension}
\usage{
rowcumsum(x)
}
\description{
Used in 'util_lprobsargsyx()'.
}
\keyword{internal}
