% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcmcstop.R
\name{funMCSELD}
\alias{funMCSELD}
\title{Calculate MC standard error using LaplacesDemon's batch means}
\usage{
funMCSELD(x)
}
\arguments{
\item{x}{matrix, each row being a "trace", that is a set of MC samples, whose MCSE is to be estimated.}
}
\value{
MCSE estimates, one for each trace. Division by sqrt(N) is already performed.
}
\description{
This function gives a good approximation of the "true" standard deviation in the case of independent samples. Multiply by \code{qnorm(x)} to obtain the \code{x}-quantile.
}
\details{
Modified from \url{https://github.com/LaplacesDemonR/LaplacesDemon/blob/master/R/ESS.R}.

Tested also on t-distributions with df=1.1 and Pareto with a=1.5 (mean exists, variance infinite).

\verb{sd() / sqrt(funESS3()} gives essentially identical results to \code{funMCSELD()}, but it's 20 times slower.

Used in 'util_combineYX()' in 'Pr()'.
}
\keyword{internal}
