% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_buildauxmetadata.R
\name{buildauxmetadata}
\alias{buildauxmetadata}
\title{Build augmented metadata file}
\usage{
buildauxmetadata(data, metadata, Dthreshold = 1, tscalefactor = 4.266)
}
\arguments{
\item{data}{data.frame object}

\item{metadata}{data.frame object}

\item{Dthreshold}{Positive number: threshold of fraction
of unique datapoints to total datapoints, to decide
whether to treat a rounded variate as continuous}

\item{tscalefactor}{Positive number: scaling factor for variate conversion}
}
\value{
A \link[base:data.frame]{data frame} with auxmetadata.
}
\description{
Builds an augmented metadata \link[base:data.frame]{data frame} from the metadata and data given to 'learn()'. This augmented metadata object is saved in the 'learnt' object produced by 'learn()'.
}
\details{
In addition to the original metadata it contains info about transformed variates and their domains, estimated location- and scale-parameters, and similar metadata.

Used in 'learn()'.
}
\keyword{internal}
