% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{metadataExample}
\alias{metadataExample}
\title{Example metadata file}
\format{
\subsection{\code{metadataExample}}{

A \link[base:data.frame]{data frame} with 2 rows and 10 columns.
}
}
\usage{
metadataExample
}
\value{
No return value.
}
\description{
A \link[base:data.frame]{data frame} containing the prior information about the variates \code{species} and \code{bill_len} of the \link[datasets:penguins]{datasets::penguins} dataset.
}
\seealso{
\code{\link[=metadatatemplate]{metadatatemplate()}} which helps producing this kind of metadata files from a given dataset.

\code{\link[=learn]{learn()}} which needs this kind of metadata files to "learn" from data.
}
\keyword{datasets}
