% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_plotFsamples.R
\name{plotFsamples}
\alias{plotFsamples}
\title{Plot one-dimensional posterior probabilities}
\usage{
plotFsamples(
  filename,
  learnt,
  data,
  plotprobability = TRUE,
  plotvariability = "samples",
  nFsamples = NULL,
  datahistogram = !(missing(data) || is.null(data)),
  datascatter = !(missing(data) || is.null(data)),
  parallel = TRUE
)
}
\arguments{
\item{filename}{Character: name of plot output file}

\item{learnt}{Either a character with the name of a directory or full
path for an 'learnt.rds' object, or such an object itself}

\item{data}{data.table object or filepath: datapoints}

\item{plotprobability}{Logical: plot the resulting probability curve}

\item{plotvariability}{Character, either 'samples' or 'quantiles':
how to plot the variability of the probability distribution with new samples}

\item{nFsamples}{Positive number: if plotvariability='samples', then
number of samples of representative frequency distributions to display
as variability;
if plotvariability='quantiles', then
the quantiles (in range 0 to 0.5) to show}

\item{datahistogram}{Logical: plot the data as histogram?}

\item{datascatter}{Logical: plot the data as scatterplot along the x-axis?}

\item{parallel}{Logical or numeric: whether to use pre-existing parallel
workers, or how many to create and use}
}
\value{
No return value.
}
\description{
Used in 'learn()' to plot diagnostics.
}
\keyword{internal}
