% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcmcstop.R
\name{fftNGS}
\alias{fftNGS}
\title{Find optimal FFT size}
\usage{
fftNGS(N)
}
\arguments{
\item{N}{Integer.}
}
\value{
Optimal FFT size
}
\description{
Modified from rstan \url{https://github.com/stan-dev/rstan/blob/develop/rstan/rstan/R/monitor.R}.
}
\details{
Used in 'funAC()'.
}
\keyword{internal}
