% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{rowinvcumsum}
\alias{rowinvcumsum}
\title{Inverse cumulative sum along first dimension}
\usage{
rowinvcumsum(x)
}
\description{
Used in 'util_lprobsargsyx()'.
}
\keyword{internal}
