% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcsamples.R
\name{mcsubset}
\alias{mcsubset}
\title{Eliminate samples from mcsamples object}
\usage{
mcsubset(learnt, subsamples)
}
\description{
Used in 'learn()'.
}
\keyword{internal}
