% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcsamples.R
\name{mcjoin}
\alias{mcjoin}
\title{Concatenate mcsample objects}
\usage{
mcjoin(x, y)
}
\description{
Used in 'learn()'.
}
\keyword{internal}
