% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcmcstop.R
\name{funMCEQ}
\alias{funMCEQ}
\title{Calculate credibility quantiles on estimated quantile}
\usage{
funMCEQ(x, prob = c(0.055, 0.945), Qpair = pnorm(c(-1, 1)))
}
\arguments{
\item{x}{A vector of MC samples}

\item{prob}{numeric vector of probabilities: quantiles whose error interval is being estimated.}

\item{Qpair}{vector of length two (further elements are ignored): lower and higher credibility-quantiles requested. Default yields a credibility interval of 68\%, or one nominal normal standard deviation.}
}
\value{
A matrix with two rows and as many columns as elements in 'prob'. Forr each column, the first and second row determine the lower and upper bound of the credibility interval of width \code{Qpair[2] - Qpair[2]}.
}
\description{
Calculates the lower and upper bound of a credibility interval, for various quantiles of the empirical distribution of a vector of MC samples.
}
\details{
Tests show that it gives a very good approximation of the "true" credibility quantiles in the case of independent samples.

Tested also on t-distributions with df=1.1 and Pareto with a=1.5 (mean exists, variance infinite).

Used in 'workerfun()' in 'learn()'
}
\keyword{internal}
